
import java.util.ArrayList;

/**
 * A parallel algorithm for an euclidean 2d TSP Problem
 */

public class ParLocalTSP implements TSPInf {

    protected Graph graph;
    protected Point[] points;
    protected Path bestPath; // beware of threads
    protected BestLocStore best;

    protected int number = 0;

    public ParLocalTSP(Point[] p, int th) {
        this.points = p;
        graph = new PlaneGraph(p);
        number = th;
    }


    public Path getBest() {
        return getBest(Long.MAX_VALUE);
    }

    public Path getBest(long max) {
        bestPath = PathArrayList.standardPath(graph); // standard path
        if ( bestPath.length() <= 2 ) {
            return bestPath;
        }
        best = new BestLocStore( bestPath, 0, number );
        Path path = PathArrayList.emptyPath(graph); // empty path
        Deque globalStack = new Deque(10000);
        try {
            globalStack.push( path );
        } catch (InterruptedException e) {
            e.printStackTrace();
        }
        Thread[] t = new Thread[number];
        for ( int i = 0; i < number; i++ ) {
            t[i] = new RunLocTSP(globalStack,best,max);
            t[i].start();
        }
        try {
            for ( int i = 0; i < number; i++ ) {
                t[i].join();
            }
        } catch (InterruptedException e) {
            e.printStackTrace();
        }
        return best.getPath();
    }

}

class BestLocStore {

    private Path path;
    private int idlers;
    public final int threads;

    public BestLocStore(Path b, int i, int t) {
        path = b;
        idlers = i;
        threads = t;
    }

    public Path getPath() {
        return path;
    }

    public void setPath(Path p) {
        path = p;
    }

    public synchronized void incIdle() {
        idlers++;
    }

    public synchronized void decIdle() {
        idlers--;
    }

    public synchronized boolean allIdle() {
        return (idlers == threads);
    }
}


class RunLocTSP extends Thread {

    private BestLocStore best;
    private Deque globalStack;
    private long maxiter;
    private long iter;
    private long pushes;
    private ArrayList stack;
    public final int COUNT;

    public RunLocTSP(Deque s, BestLocStore b, long max) {
        globalStack = s;
        best = b;
        COUNT = b.getPath().maxsize()+5; //*2 / +5 ?
        maxiter = max;
        stack = new ArrayList(500);
    }


    public void run() {
        iter = 0;
        pushes = 0;
        boolean running = true;
        while (running) {
            Object o = null;
            if ( stack.isEmpty() ) {
               try {
                  best.incIdle();
                  if ( globalStack.empty() && best.allIdle() ) {
                      for ( int i = 0; i < best.threads-1; i++) {
                          globalStack.push(null);
                      }
                      best.decIdle();
                      break; //return;
                  }
                  o = globalStack.pop();
                  best.decIdle();
               } catch (InterruptedException e) {
                  e.printStackTrace();
                  running = false;
               }
            } else {
                o = stack.remove( stack.size()-1 );
            }
            if ( o == null ) {
                break; //return;
            }
            Path path = (Path)o;
            getBest( path );
        }
        System.out.println("iterations = " + iter + " pushes " + pushes);
    }


    public void getBest(Path path) {
        Path[] ps = path.nextPaths(); 
        iter += ps.length;
        double bestCost = best.getPath().cost();
        if ( path.length()+1 == path.maxsize() ) {
           for ( int i = 0; i < ps.length; i++ ) { // branch
               if ( ps[i].cost() < bestCost ) {
                 synchronized ( best ) {
                     if ( ps[i].cost() < best.getPath().cost() ) {
                        bestCost = ps[i].cost();
                        best.setPath( ps[i] );
                        System.out.println("par best = " + ps[i]);
                   }
                 }
               }
           }
           return;
        }
        if ( iter >= maxiter ) {
            return;
        }
        for ( int i = 0; i < ps.length; i++ ) { // branch
            if ( ps[i].cost() < bestCost ) { // cut
               // getBest( ps[i] ); // changes best and bestPath
               stack.add( ps[i] );
            }
        }
        if ( stack.size() > COUNT && path.length()+6 <= path.maxsize() ) {
            if ( globalStack.empty() ) {
               Object o = stack.remove( stack.size()-1 ); // 0 breadth first
               pushes++;
               try {
                   globalStack.push( o ); // beware of deadlock
               } catch (InterruptedException e) {
                   e.printStackTrace();
               }
            }
        }
        return;
    }

}